///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoBridge.cc
// --------------
// CegoBridge client program
//
// Design and Implementation by Bjoern Lemke               
//                                                         
// (C)opyright 2010-2015 Bjoern Lemke                        
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoBridge
// 
// Description: 
//
//
///////////////////////////////////////////////////////////////////////////////

#include <locale.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <readline/readline.h>
#include <readline/history.h>

#include <config.h>

#include <lfcbase/Chain.h>
#include <lfcbase/Exception.h>
#include <lfcbase/GetLongOpt.h>
#include <lfcbase/File.h>
#include <lfcbase/Tokenizer.h>
#include <cego/CegoNet.h>

#include "MySQLAction.h"
#include "MySQLFormater.h"

#include "OracleAction.h"
#include "OracleFormater.h"

#include "PostgresAction.h"
#include "PostgresFormater.h"


#define CEGOBRIDGE_PRODUCT PACKAGE_NAME
#define CEGOBRIDGE_VERSION PACKAGE_VERSION
#define CEGOBRIDGE_COPYRIGHT "Copyright (C) 2010-2018 by Bjoern Lemke. All rights reserved"

#define USAGE "Usage: cegobridge --user=<user>/<password>\n\
          --tableset=<tableset>\n\
          --dbtype=ora|mysql|postgres\n\
          --mode=import|export\n\
          [ --sqlfile=<filename> ]\n\
          [ --server=<host>]\n\
          [ --port=<intval> ]\n\
          [ --protocol=(serial|fastserial|xml)]\n\
          [ --logfile=<filename> ]\n\
          [ --nocasesensitive ]\n\
          [ --maxsendlen=<intval> ]\n\
          [ --dump ]\n\
          [ --debug ] [ --version  ] [ --help ]"

#define DEFAULTPORT 2200
#define DEFAULTSERVER "localhost"
#define DEFAULTMAXSENDLEN 8192
#define DEFAULTPROTOCOL "serial"
#define MYSQL_DB "mysql"
#define ORACLE_DB "oracle"
#define POSTGRES_DB "postgres"

#define IMPORT_MODE "import"
#define EXPORT_MODE "export"

// size info buffer len
#define SIZEBUFLEN 10
#define MSGBUFLEN 500 

#define LANGUAGE_ENV "LANG"
#define LOCALE_CATEGORY LC_TIME

extern char __lfcVersionString[];
extern char __lfcxmlVersionString[];


int main(int argc, char **argv)
{	
    
    GetLongOpt longOpt(argc, argv);
    
    longOpt.addOpt("version");
    longOpt.addOpt("help");
    longOpt.addOpt("logfile");
    longOpt.addOpt("tableset");
    longOpt.addOpt("mode");
    longOpt.addOpt("user");
    longOpt.addOpt("dbtype");
    longOpt.addOpt("mode");
    longOpt.addOpt("sqlfile");
    longOpt.addOpt("debug");
    longOpt.addOpt("server", DEFAULTSERVER);
    longOpt.addOpt("port", Chain(DEFAULTPORT));
    longOpt.addOpt("maxsendlen", DEFAULTMAXSENDLEN);
    longOpt.addOpt("nocasesensitive");
    longOpt.addOpt("dump");
    longOpt.addOpt("protocol", Chain(DEFAULTPROTOCOL));
    
    try
    {
	longOpt.parseOpt(); 
    }
    catch ( Exception e )
    {
	Chain msg;
	e.pop(msg);
	cerr << msg << endl;
	cerr << USAGE << endl;
	exit(1);	
    }
    
    Chain user;
    Chain password;
    
    try
    {
    
	bool debug=false;
	if ( longOpt.isSet("debug") )
	    debug=true;
	
	bool isCaseSensitive = true;
	if ( longOpt.isSet("nocasesensitive") )
	    isCaseSensitive=false;
	
	bool doDump = false;
	if ( longOpt.isSet("dump") )
	    doDump=true;
	
	if ( longOpt.isSet( Chain("help") ) )
	{	
	    cerr << USAGE << endl;
	    exit(0);
	}
	
	if ( longOpt.isSet( Chain("version") ) )
	{
	    cout << CEGOBRIDGE_PRODUCT << " (" << sizeof(long) * 8 << " bit), Version " << CEGOBRIDGE_VERSION 
		 << " [ lfc : " << __lfcVersionString  
		 << ", lfcxml : " <<  __lfcxmlVersionString << " ]" << endl;
	    cout << CEGOBRIDGE_COPYRIGHT << endl;
	    exit(0);
	}

	Chain prot = longOpt.getOptValue("protocol");
	CegoDbHandler::ProtocolType protType;
	
	if ( prot == Chain("serial") )
	{
	    protType = CegoDbHandler::SERIAL;
	}
	else if ( prot == Chain("fastserial") )
	{
	    protType = CegoDbHandler::FASTSERIAL;
	}
	else if ( prot == Chain("xml") )
	{
	    protType = CegoDbHandler::XML;
	}
	else
	{
	    cerr << "Invalid protocol " << prot;
	    exit (1);
	}
	
	Chain logFile = longOpt.getOptValue("logfile");
	Chain serverName = longOpt.getOptValue("server");
	Chain tableSet = longOpt.getOptValue("tableset");
	int portNo = longOpt.getOptValue("port").asInteger();
	int maxSendLen = longOpt.getOptValue("maxsendlen").asInteger();
	
	if ( doDump == false )
	{
	    Chain authString = longOpt.getOptValue("user");
	    
	    Tokenizer authTok(authString, Chain("/")); 
	    authTok.nextToken(user);
	    authTok.nextToken(password);
	    
	    if ( user.length() == 0 )
	    {
		cerr << "User not set" << endl;
		exit (1);
	    }
	    if ( password.length() == 0 )
	    {
		cerr << "Password not set" << endl;
		exit (1);
	    }
	    if ( tableSet.length() == 0 )
	    {
		cerr << "Tableset not set" << endl;
		exit (1);
	    }
	    
	    // for localization
	    char* lang = 0;
	    if ( ( lang = getenv(LANGUAGE_ENV) ) != 0)
	    {
		if ( setlocale(LOCALE_CATEGORY, lang) == 0)
		{
		    Chain msg = Chain("Cannot set locale ") + Chain(lang);
		    cerr << msg << endl;
		    exit(1);
		}
	    }
	}
	
	int lineNo=0;
	
	CegoNet *pCegoNet = 0;

	Chain logMode("notice");
	if ( debug )
	    logMode=Chain("debug");

	if ( doDump == false )
	{
	    pCegoNet = new CegoNet( protType, logFile, Chain(""), logMode );
	    pCegoNet->connect(serverName, portNo, tableSet, user, password, maxSendLen);
	    pCegoNet->doQuery("set append on;");
	}
	
	Chain dbtype = longOpt.getOptValue("dbtype");
	Chain mode = longOpt.getOptValue("mode");
	Chain sqlfile = longOpt.getOptValue("sqlfile");

	if ( mode == Chain(IMPORT_MODE) )
	{	       

	    if ( dbtype == Chain(MYSQL_DB) )
	    {
		MySQLAction pa(sqlfile, pCegoNet, isCaseSensitive);		
		pa.import();
	    }
	    else if ( dbtype == Chain(ORACLE_DB) )
	    {
		OracleAction pa(sqlfile, pCegoNet, isCaseSensitive);		
		pa.parse();		
	    }
	    else if ( dbtype == Chain(POSTGRES_DB) )
	    {
		PostgresAction pa(sqlfile, pCegoNet, isCaseSensitive);		
		pa.parse();		
	    }
	}
	else
	{	

	    if ( dbtype == Chain(MYSQL_DB) )
	    {
		cout << "MySQL Export" << endl;
		cout << "------------" << endl;
		
		MySQLFormater mysqlFormater(pCegoNet, sqlfile);

		cout << "Exporting tables ..." << endl;
		mysqlFormater.exportTable();

		cout << "Exporting procedures ..." << endl;
		mysqlFormater.exportProcedure();

		cout << "Exporting views ..." << endl;
		mysqlFormater.exportView();

		cout << "Exporting indexes ..." << endl;
		mysqlFormater.exportIndex();
		mysqlFormater.exportBtree();

		cout << "Exporting constraints ..." << endl;
		mysqlFormater.exportConstraint();		
	    }
	    else if ( dbtype == Chain(ORACLE_DB) )
	    {
		cout << "Oracle Export" << endl;
		cout << "-------------" << endl;

		OracleFormater oracleFormater(pCegoNet, sqlfile);
		cout << "Still not implemented" << endl;
	    }
	    else if ( dbtype == Chain(POSTGRES_DB) )
	    {
		cout << "Postgres Export" << endl;
		cout << "---------------" << endl;

		PostgresFormater postgresFormater(pCegoNet, sqlfile);
		cout << "Still not implemented" << endl;
	    }
	}
		
	if ( pCegoNet )
	{
	    pCegoNet->disconnect();
	    delete pCegoNet;
	}
    }
    catch ( Exception e )
    {
	Chain msg;
	e.pop(msg);
	cout << msg << endl;
	exit(1);
    }

    exit(0);
}


