///////////////////////////////////////////////////////////////////////////////
//                                                         
// MySQLAction.cc  
// --------------                                                        
// MySQL semantic parser actions
//
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2015,2016,2017 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: MySQLAction
// 
// Description: 
//
//
///////////////////////////////////////////////////////////////////////////////


#include <ctype.h>
#include <stdlib.h>

#include "MySQLAction.h"

#define MYSQL_MAXLINELEN 10000000
#define MAXCGSTRLEN 5000
#define DELIMITER_TOKEN "delimiter"
#define DELIMITER ";"
#define MAXINSERT 100

MySQLAction::MySQLAction(const Chain& fileName, CegoNet *pCegoNet, bool isCaseSensitive)
{
    if ( fileName.length() > 0 )
    {
	_pFile = new File(fileName);   
	_pFile->open(File::READ);
    }
    else
    {
	_pFile = new File(File::IN);   
    }
    _i = 0;
    _pC = 0;
    _pCegoNet = pCegoNet;
    _isFirst=true;
    _isCaseSensitive = isCaseSensitive;
    _delimiter = Chain(DELIMITER);
}

MySQLAction::~MySQLAction()  
{
}


void MySQLAction::import()
{
    try
    {
	// for dump mode, we enable append mode
	if ( _pCegoNet == 0)
	    cout << "set append on;" << endl;

	long s=0;
	while ( readStatement() )
	{	    
	    // cout << "Parsing stmt " << s << " of length " << _line.length() << endl;
	    parse();
	    s++;
	}
    }
    catch ( Exception e )
    {
	Chain msg;
	e.pop(msg);
	
	Chain fullmsg = Chain("Parse Error at line :") + _line + Chain("Error = ") + msg;
	
	throw Exception(EXLOC, fullmsg);
    }
}

bool MySQLAction::readStatement()
{
    _line = Chain("");
    
    while ( 1 )
    {
	Chain inputLine;
	if ( _pFile->readLine(inputLine, MYSQL_MAXLINELEN) > 0 )
	{	  
	    if ( inputLine.length() >= 2 )
	    {
		Chain checkCommentAndSet = inputLine.cutTrailing(" \t");
		if (checkCommentAndSet.subChain(1, 2) != Chain("--") && checkCommentAndSet.subChain(1, 3) != Chain("SET"))
		{
		    int pos;

		    // cout << "Checking delimiter " << inputLine.subChain(1, Chain(DELIMITER_TOKEN).length()).toLower() << endl;

		    
		    if ( inputLine.length() >= Chain(DELIMITER_TOKEN).length() && inputLine.subChain(1, Chain(DELIMITER_TOKEN).length()-1).toLower() == Chain(DELIMITER_TOKEN) )
		    {
			_delimiter = inputLine.subChain(pos + Chain(DELIMITER_TOKEN).length(), inputLine.length()).cutTrailing(" \t");
			// cout << "Set delimiter to <" << _delimiter << ">" << endl;
		    }		 
		    else if ( inputLine.subChain(inputLine.length() - _delimiter.length() + 1, inputLine.length()) == _delimiter )
		    {
			// cout << "Input = " << inputLine << endl;
			// cout << "Found delimiter" << endl;
			
			_line += inputLine.truncRight(_delimiter);
			
			 // cout << " Del = " << _delimiter << " LINE = " << _line << endl;
			_pC = (char*)_line;
			_i=0;
			return true;
		    }
		    else
		    {

			// cout << "Delimiter is " << _delimiter << endl;
			// cout << "Delimiter match = <" << inputLine.subChain(inputLine.length() - _delimiter.length() + 1, inputLine.length()) << ">" << endl;
			// cout << "Concat line .." << endl;
			_line += inputLine;
			// force separator
			_line += Chain(" ");
		    }			    
		}
	    }
	}
	else
	{
	    return false;
	}
    }
}

char MySQLAction::nextChar() 
{   
   
    if ( *_pC == '\'')
    {
	_pC++;
	readChain();
	return 0;
    }

    while ( *_pC == '/' && *(_pC+1) == '*' )
    {
	_pC++;
	_pC++;

	bool eoc=false;
	while ( ! eoc )
	{
	    while ( *_pC != '*' )
	    {
		_pC++;
		if ( *_pC == 0 ) 
		{
		    return 0; 
		}
	    }
	    if ( *_pC == '*' && *(_pC+1) == '/' )
	    {
		eoc=true;
		_pC++;
	    }	   
	    _pC++;	    
	}	
    }

    if ( *_pC == 0 ) 
    {
	return 0; 
    }
        
    char c = *_pC; 
    _pC++; 

    if ( _isCaseSensitive )
	return c;
    else
	return (char)tolower((int)c);
} 

void MySQLAction::readChain()
{

    setReserved(STRINGVAL); 
    
    int i=0;    

    
    while ( *_pC != '\'') 
    { 
	// cout << "Storing: " << *_pC << endl; 
	
	if ( *_pC == '\\') 
	{ 
	    _pC++; 
	    
	    if ( *_pC == 'n') 
	    { 
		_stringBuf[i] = '\\'; 
                i++;
		_stringBuf[i] = 'n'; 
	    } 
	    else if ( *_pC == 't') 
	    { 
		_stringBuf[i] = '\t'; 
                i++;
		_stringBuf[i] = 't'; 
	    } 
	    else if ( *_pC == 'r') 
	    { 
		_stringBuf[i] = '\r'; 
                i++;
		_stringBuf[i] = 'r'; 
	    } 
	    else if ( *_pC == '0') 
	    { 
		_stringBuf[i] = '\0'; 
	    }
	    else if ( *_pC == '"') 
	    { 
		_stringBuf[i] = '"'; 
	    } 
	    else if ( *_pC == '\'') 
	    { 
		_stringBuf[i] = '\\'; 
                i++;
		_stringBuf[i] = '\''; 
	    }

	    else if ( *_pC == '\\') 
	    { 
		_stringBuf[i] = '\\'; 
                i++;
		_stringBuf[i] = '\\'; 
	    } 
	    else 
	    { 
		Chain msg = Chain("Invalid escape character <") + Chain(*_pC) + Chain(">"); 
		throw Exception(EXLOC, msg);
	    } 
	} 
	else
	{ 
	    _stringBuf[i] = *_pC; 
	} 
	
	i++; 
	
	if (i == MYSQL_MAXSTRINGLEN) 
	{ 
	    Chain msg = Chain("Stringbuf exceeded (") + Chain(MYSQL_MAXSTRINGLEN) + Chain(")"); 
	    throw Exception(EXLOC, msg);
	}
	
	_pC++;
	
	if ( *_pC == 0 )
	{
	    Chain msg("Unterminated string"); 
	    throw Exception(EXLOC, msg);
	}
    }
    
    _pC++; 

    _stringBufLen = i+1;
    _stringBuf[i] = 0; 

    return; 
}

void MySQLAction::backChar() 
{
    if ( _pC != 0 )
	_pC--;
}
    
void MySQLAction::printTokenList() 
{    
    cout << "TokenList is " << endl;
	
    ListT<Chain> tlist =  getTokenList();
    Chain* pS = tlist.First();
    while (pS)
    {
	cout << *pS << endl;
	pS = tlist.Next();
    }    
}

void MySQLAction::cleanUp() 
{
    
}

void MySQLAction::execCreate() 
{

    Chain stmt = Chain("create table ") + _table + Chain(" (");

    Chain *pCol = _colList.First();
    while ( pCol )
    {
	stmt += Chain(" ") + *pCol;
	pCol = _colList.Next();
	if ( pCol )
	    stmt += Chain(",");
	
    }
    stmt += Chain(");");
    _colList.Empty();   


    // cout << "Create = " << stmt << endl;
    if ( _pCegoNet )
    {
	cout << "Creating table " << _table << endl;
	_pCegoNet->doQuery(stmt);
    }
    else
	cout << stmt << endl;


    Chain *pStmt = _postCreateStmtList.First();
    while ( pStmt )
    {
	if ( _pCegoNet )
	{
	    cout << ".";
	    cout.flush();
	    _pCegoNet->doQuery(*pStmt);
	}
	else
	    cout << *pStmt << endl;
	pStmt = _postCreateStmtList.Next();
    }

    _postCreateStmtList.Empty();
    
}

void MySQLAction::insertPreAction()
{
    if ( _pCegoNet )
    {
	cout << "Inserting into " << _table << " ";
	cout.flush();
    }
} 

void MySQLAction::insertPostAction()
{
    if ( _pCegoNet )
    {
	cout << endl;
    }
}

void MySQLAction::insertAttr()
{
}

void MySQLAction::storeInsert()
{
   _valArray.Insert(_valList);
   _valList.Empty();
}
   
void MySQLAction::execInsert() 
{
    ListT<Chain>* pValList = _valArray.First();

    bool nextStmt=true;
    int numInsert=0;
    
    Chain stmt;

    while ( pValList )
    {

	if ( numInsert == 0 )
	    stmt = Chain("insert into ") + _table + Chain(" values ");
      
	stmt += Chain("(");
	
	Chain *pVal = pValList->First();
	int pos=0;
	while ( pVal )
	{
	    TypeInfo* pTI = getTypeInfo(_table, pos);
	    if ( pTI == 0 )
		throw Exception(EXLOC, Chain("No type info for table ") + _table + Chain("Position ") + Chain(pos) );
	    
	    if ( pTI->getType() == Chain("datetime") && *pVal != Chain("null") )
	    {
		if ( pVal->subChain(2,5) == Chain("1899") )
		    stmt += Chain(" null");	    
		else if ( pVal->subChain(2,11) == Chain("0000-00-00") )
		    stmt += Chain(" date('%Y-%m-%d %H:%M:%S','1970-01-01 00:00:00')");
		else
		    stmt += Chain(" date('%Y-%m-%d %H:%M:%S',") + *pVal + Chain(")");
	    }
	    else if ( pTI->getType() == Chain("date") && *pVal != Chain("null") )
	    {
		if ( pVal->subChain(2,5) == Chain("1899") )
		    stmt += Chain(" null");
		else if ( pVal->subChain(2,11) == Chain("0000-00-00") )
		    stmt += Chain(" date('%Y-%m-%d','1970-01-01')");	    
		else
		    stmt += Chain(" date('%Y-%m-%d',") + *pVal + Chain(")");
	    }
	    else
		if ( pVal->length() > MAXCGSTRLEN )
		    stmt += Chain("'") + pVal->subChain(1, MAXCGSTRLEN) + Chain("'");
		else
		    stmt += Chain(" ") + *pVal;
	    
	    
	    pVal = pValList->Next();
	    if ( pVal )
		stmt += Chain(",");
	    
	    pos++;
	}
	stmt += Chain(")");
	pValList = _valArray.Next();
	if ( pValList && numInsert < MAXINSERT - 1 )
	{
	    stmt += Chain(",");
	    numInsert++;
	}
	else
	{
	    stmt += Chain(";");
	    if ( _pCegoNet )
	    {
		cout << ".";
		cout.flush();
		_pCegoNet->doQuery(stmt);
	    }
	    else
		cout << stmt << endl;
	    numInsert=0;
	}       
    }
    _valArray.Empty();
}

void MySQLAction::storeTable() 
{
    Chain* pS = getTokenList().First();
    _table = pS->cutTrailing("`");
}

void MySQLAction::execDrop()
{
    Chain stmt = Chain("drop if exists table ") + _table +  Chain(";");

    if ( _pCegoNet )
	_pCegoNet->doQuery(stmt); 
    else
	cout << stmt << endl;

}

void MySQLAction::dropOpt1()
{    
}

void MySQLAction::dropOpt2()
{
}

void MySQLAction::storeAttrColumn()
{
    Chain* pS = getTokenList().First();
    
    Chain attrName =  pS->cutTrailing("`");

    if ( attrName == Chain("sync") )
	attrName = Chain("a_sync");
    else if ( attrName == Chain("start") )
	attrName = Chain("a_start");

    Chain typeString=_dataTypeString;

    // in case of date type, we normalize to datetime
    if ( _dataTypeString == Chain("date") )
	typeString=Chain("datetime");
    
    Chain colString = attrName + Chain(" ") + typeString;
    
    if ( _defaultOptString.length() > 1 )
    {

	/* in case of strange default date values, we have to convert */ 
	if ( typeString == Chain("datetime") && _defaultOptString == Chain("'0000-00-00'") )
	    _defaultOptString = Chain("''");
	

	colString += Chain(" default ") + _defaultOptString;
	if ( _defaultOptString == Chain("''") )
	    _nullOptString = "";
    }
    
    if ( _nullOptString.length() > 1 )
	colString += Chain(" ") + _nullOptString;
    
    // cout << "ColString = " << colString << endl;

    putTypeInfo( TypeInfo(_table, _colList.Size(), _dataTypeString) );
    _colList.Insert(colString);
}

void MySQLAction::storePrimaryKey()
{
    Chain stmt = Chain("create primary btree on ") + _table + Chain("(");

    Chain *pKeyCol = _keyColList.First();
    while ( pKeyCol )
    {
	stmt += Chain(" ") + *pKeyCol;
	pKeyCol = _keyColList.Next();
	if ( pKeyCol )
	    stmt += Chain(",");	
    }
    stmt += Chain(");");

    _keyColList.Empty();

    _postCreateStmtList.Insert(stmt);
}

void MySQLAction::storeUniqueKey()
{

    Chain* pS = getTokenList().First();
    pS = getTokenList().Next();
    pS = getTokenList().Next();

    Chain btreeName = _table + Chain("_") + pS->cutTrailing("`");
    
    Chain stmt = Chain("create unique btree ") + btreeName + Chain(" on ") + _table + Chain("(");

    Chain *pKeyCol = _keyColList.First();
    while ( pKeyCol )
    {
	stmt += Chain(" ") + *pKeyCol;
	pKeyCol = _keyColList.Next();
	if ( pKeyCol )
	    stmt += Chain(",");	
    }
    stmt += Chain(");");

    _keyColList.Empty();

    _postCreateStmtList.Insert(stmt);
}

void MySQLAction::storeKey()
{
    Chain* pS = getTokenList().First();
    pS = getTokenList().Next();
    pS = getTokenList().Next();

    // mysql allows non unique btree names, so we have so make unique by using tablename as prefix 
    Chain btreeName = _table + Chain("_") + pS->cutTrailing("`");

    Chain stmt = Chain("create btree ") + btreeName + Chain(" on ") + _table + Chain("(");

    Chain *pKeyCol = _keyColList.First();
    while ( pKeyCol )
    {
	stmt += Chain(" ") + *pKeyCol;
	pKeyCol = _keyColList.Next();
	if ( pKeyCol )
	    stmt += Chain(",");	
    }
    stmt += Chain(");");

    _keyColList.Empty();

    _postCreateStmtList.Insert(stmt);
}



void MySQLAction::storeKeyColumn()
{
    Chain* pS = getTokenList().First();
    _keyColList.Insert(pS->cutTrailing("`"));
}

void MySQLAction::storeNullOpt()
{
    _nullOptString = "";
}

void MySQLAction::storeNotNullOpt()
{
    _nullOptString = "not null";
}

void MySQLAction::storeDefaultOpt()
{
    _defaultOptString = _constValue;
}

void MySQLAction::storeDefaultNullOpt()
{
    _defaultOptString = "";
}

void MySQLAction::insertConstValue()
{
    _valList.Insert(_constValue);
}

void MySQLAction::storeIntType()
{
    _dataTypeString = Chain("int");
}

void MySQLAction::storeLongType()
{
    _dataTypeString = Chain("long");
}

void MySQLAction::storeDecimalType()
{
    _dataTypeString = Chain("fixed(") + Chain(_decDim) + Chain(")");
}

void MySQLAction::storeUnsignedType()
{
}

void MySQLAction::storeVarcharType()
{
    Chain* pS = getTokenList().First();
    pS = getTokenList().Next();

    int vcDim = pS->asInteger();

    // since MySQl stores values varchars as multibyte, we have to double the array dimension
    vcDim=2*vcDim;
    
    _dataTypeString = Chain("string(") + Chain(vcDim) + Chain(")");
}

void MySQLAction::storeTimestampType()
{
    _dataTypeString = Chain("datetime");
}

void MySQLAction::storeDatetimeType()
{
    _dataTypeString = Chain("datetime");
}

void MySQLAction::storeDateType()
{
    _dataTypeString = Chain("date");
}

void MySQLAction::storeMediumTextType()
{
    _dataTypeString = Chain("string(") + Chain(MAXCGSTRLEN) + Chain(")");
}

void MySQLAction::storeEnumType()
{
    // for now, we map enum to string 100
    _dataTypeString = Chain("string(100)");
}


void MySQLAction::storeCurrentTimestampValue()
{
    _constValue = Chain("sysdate");
}

void MySQLAction::storeIntValue()
{
    Chain* pS = getTokenList().First();
    
    if ( pS->length() < 10 )
	_constValue = *pS;
    else
	_constValue = Chain("(long)") + *pS;
}

void MySQLAction::storeNegIntValue()
{
    Chain* pS = getTokenList().First();
    
    if ( pS->length() < 10 )
	_constValue = Chain("-") + *pS;
    else
	_constValue = Chain("(long)-") + *pS;
}

void MySQLAction::storeVarcharValue()
{
    _constValue = Chain("'") + Chain(_stringBuf) + Chain("'");
}

void MySQLAction::storeNullValue()
{
    _constValue = Chain("null");
}

void MySQLAction::storeDim()
{
    
    Chain* pS = getTokenList().First();
    pS = getTokenList().Next();
    _dim = pS->asInteger();
}

void MySQLAction::storeDecimalDim()
{
    Chain* pS = getTokenList().First();
    pS = getTokenList().Next();
    _decDim = pS->asInteger();
    pS = getTokenList().Next();
    pS = getTokenList().Next();
    _dim = pS->asInteger();
}

void MySQLAction::procStore()
{
}

void MySQLAction::procHead()
{
}

void MySQLAction::procReturnOpt()
{
}

void MySQLAction::procArg1()
{
}

void MySQLAction::procStoreBlock()
{
}

void MySQLAction::procDeclare()
{
}

void MySQLAction::procAssign()
{
}

void MySQLAction::procExpression()
{
}

void MySQLAction::procInsert()
{
}

void MySQLAction::procReturn1()
{
}

void MySQLAction::procReturn2()
{
}

void MySQLAction::procBlock()
{
}

void MySQLAction::procBlockStart()
{
}

void MySQLAction::procIf()
{
}

void MySQLAction::procStoreIfBlock()
{
}

void MySQLAction::procStoreElseBlock()
{
}

void MySQLAction::procElsePart()
{
}

void MySQLAction::procIfCondition()
{
}

void MySQLAction::procElsIfCondition()
{
}

void MySQLAction::procWhile()
{
}

void MySQLAction::procWhileCondition()
{
}

void MySQLAction::procCondition1()
{
}

void MySQLAction::procCondition2()
{
}

void MySQLAction::procCondition3()
{
}

void MySQLAction::procPredicate1()
{
}

void MySQLAction::procPredicate2()
{
}

void MySQLAction::equalComp()
{
}

void MySQLAction::notEqualComp()
{
}

void MySQLAction::lessComp()
{
}

void MySQLAction::moreComp()
{
}

void MySQLAction::lessEqualComp()
{
}

void MySQLAction::moreEqualComp()
{
}

void MySQLAction::exprPlus()
{
}

void MySQLAction::exprMinus()
{
}

void MySQLAction::exprTerm()
{
}

void MySQLAction::termStar()
{
}

void MySQLAction::termDiv()
{
}

void MySQLAction::termFactor()
{
}

void MySQLAction::factorConst()
{
}

void MySQLAction::factorIdentifier()
{
}

void MySQLAction::factorFunction()
{
}

void MySQLAction::factorExpr()
{
}

void MySQLAction::putTypeInfo(TypeInfo ti)
{
    // cout << " #### Putting type info " << ti.getTable() << " " << ti.getPos() << " " << ti.getType() << endl;
    _typeList.Insert(ti);
}

MySQLAction::TypeInfo* MySQLAction::getTypeInfo(const Chain& tableName, int pos)
{
    TypeInfo* pTI = _typeList.Find( TypeInfo(tableName, pos) );
    return pTI;
}
